// $URL: http://subversion:8080/svn/gsc/trunk/drivers/gsc_common/api/intime/os_file.c $
// $Rev: 43206 $
// $Date: 2018-07-08 17:03:14 -0500 (Sun, 08 Jul 2018) $

// INtime: API Library: source file

#include "main.h"



//*****************************************************************************
// Return Values:
//	0  = success
//	<0 = failure, value is -errno or -INtime error code
//	>0 = never returned
static int _file_alloc(os_file_t** fp)
{
	int			i;
	os_file_t*	ptr		= NULL;
	int			ret;

	for (;;)	// convience loop
	{
		if (fp == NULL)
		{
			ret	= -EINVAL;
			break;
		}

		for (i = 0; i < sizeofarray(os_api_global.files); i++)
		{
			if (os_api_global.files[i].fd == 0)
			{
				ptr	= &os_api_global.files[i];
				memset(ptr, 0, sizeof(os_file_t));
				break;
			}
		}

		if (ptr == NULL)
		{
			ret	= -ENOMEM;
			break;
		}

		// Success!
		os_api_global.fd_counter++;
		ptr->fd	= OS_FD_ENCODE(i);
		fp[0]	= ptr;
		ret		= 0;
		break;
	}

	return(ret);
}



//*****************************************************************************
// Return Values:
//	0  = success
//	<0 = failure, value is -errno or -INtime error code
//	>0 = never returned
int os_file_create(os_file_t** fp)
{
	os_file_t*	ptr		= NULL;
	int			ret;
	int			unlock	= 0;

	for (;;)	// convience loop
	{
		// Perform basic validation.

		if (fp == NULL)
		{
			ret	= -EINVAL;
			break;
		}

		fp[0]	= NULL;

		// Acquire exclusive global access.
		ret	= os_sem_lock(&os_api_global.sem);

		if (ret)
		{
			// We can't get exclusive access.
			break;
		}

		unlock	= 1;

		// Acquire a file structure.
		ret	= _file_alloc(&ptr);

		if ((ret) || (ptr == NULL))
		{
			ret	= -ENOMEM;
			break;
		}

		// Create the semaphore for the file structure.
		ret	= os_sem_create(&ptr->sem);

		if (ret)
			break;

		ret	= os_sem_lock(&ptr->sem);

		if (ret)
			break;

		// Success!
		fp[0]	= ptr;
		ret		= 0;
		break;
	}

	if (unlock)
		os_sem_unlock(&os_api_global.sem);

	return(ret);
}



//*****************************************************************************
// Return Values:
//	0  = success
//	<0 = failure, value is -errno or -INtime error code
//	>0 = never returned
int os_file_destroy(os_file_t** fp)
{
	int	ret;
	int	unlock	= 0;

	for (;;)	// convience loop
	{
		// Perform basic validation.

		if ((fp == NULL) || (fp[0] == NULL))
		{
			// There is nothing to do.
			ret	= -EINVAL;
			break;
		}

		// Acquire exclusive global access.
		ret	= os_sem_lock(&os_api_global.sem);

		if (ret)
		{
			// We can't get exclusive access.
			break;
		}

		os_sem_destroy(&fp[0]->sem);
		memset(fp[0], 0, sizeof(os_file_t));
		fp[0]	= NULL;
		ret		= 0;
		unlock	= 1;
		break;
	}

	if (unlock)
		os_sem_unlock(&os_api_global.sem);

	return(ret);
}



//*****************************************************************************
// Return Values:
//	0  = success
//	<0 = failure, value is -errno or -INtime error code
//	>0 = never returned
int os_file_lock(int fd, os_file_t** fp)
{
	int			i;
	os_file_t*	ptr		= NULL;
	int			ret;
	int			unlock	= 0;

	for (;;)	// convience loop
	{
		// Perform basic validation.
		i	= OS_FD_DECODE_INDEX(fd);

		if ((i < 0) || (i >= sizeofarray(os_api_global.files)))
		{
			ret	= -EINVAL;
			break;
		}

		if (fp == NULL)
		{
			// There is nothing to do.
			ret	= -EINVAL;
			break;
		}

		fp[0]	= NULL;

		// Get the global lock so we can access the file table.
		ret	= os_sem_lock(&os_api_global.sem);

		if (ret)
		{
			// We can't get exclusive access.
			break;
		}

		unlock	= 1;
		ptr		= &os_api_global.files[i];

		if (fd != ptr->fd)
		{
			ret	= -EINVAL;
			break;
		}

		ret	= os_sem_lock(&ptr->sem);

		if (ret)
		{
			// We can't get exclusive access.
			break;
		}

		// All went well.
		fp[0]	= ptr;
		break;
	}

	if (unlock)
		os_sem_unlock(&os_api_global.sem);

	return(ret);
}



//*****************************************************************************
// Return Values:
//	0  = success
//	<0 = failure, value is -errno or -INtime error code
//	>0 = never returned
int os_file_unlock(os_file_t** fp)
{
	int	ret;
	int	unlock	= 0;

	for (;;)	// convience loop
	{
		// Perform basic validation.

		if ((fp == NULL) || (fp[0] == NULL))
		{
			// There is nothing to do.
			ret	= -EINVAL;
			break;
		}

		// Get the global lock so we can access the file table.
		ret	= os_sem_lock(&os_api_global.sem);

		if (ret)
		{
			// We can't get exclusive access.
			break;
		}

		os_sem_unlock(&fp[0]->sem);
		fp[0]	= NULL;
		ret		= 0;
		unlock	= 1;
		break;
	}

	if (unlock)
		os_sem_unlock(&os_api_global.sem);

	return(ret);
}


