// $URL: http://subversion:8080/svn/gsc/trunk/drivers/gsc_common/api/intime/os_ioctl.c $
// $Rev: 43206 $
// $Date: 2018-07-08 17:03:14 -0500 (Sun, 08 Jul 2018) $

// INtime: API Library: source file

#include "main.h"



//*****************************************************************************
// Return Values:
//	0  = success
//	<0 = failure, value is -errno or -INtime error code
//	>0 = never returned
static int _ioctl_large(os_file_t** fp, const os_mbox_t* mbox, os_msg_t* msg, int cmd, void* buf, int bytes)
{
	RTHANDLE	mem		= BAD_RTHANDLE;
	u32			offset	= 0;
	int			ret		= 0;
	WORD		size;
	BOOLEAN		sts;

	for (;;)
	{
		// Create a memory handle for the buffer.
		ret	= os_mem_handle_create(buf, bytes, &offset, &mem);

		if (ret)
			break;

		// Initialize and send the message.
		memset(msg, 0, sizeof(os_msg_t));
		msg->status				= OS_MSG_STS_OK;
		msg->msg_type			= OS_MSG_TYPE_IOCTL_LG;
		msg->size				= OS_MSG_SIZE_IOCTL_LG;
		msg->mbox				= mbox->mbox;
		msg->u.ioctl_lg.cmd		= cmd;
		msg->u.ioctl_lg.mem		= (u32) mem;
		msg->u.ioctl_lg.offset	= offset;
		sts						= SendRtData(fp[0]->dev_mbox, msg, OS_MSG_SIZE_IOCTL_LG);

		if (sts == FALSE)
		{
			ret	= os_get_error_status();
			break;
		}

		// Release the file pointer and don't use it after this point.
		os_file_unlock(fp);

		// Get and process the response.
		size	= ReceiveRtData(mbox->mbox, msg, WAIT_FOREVER);

		if (size == 0)
		{
			// The operation failed.
			ret	= os_get_error_status();
			break;
		}

		if ((size != OS_MSG_SIZE_IOCTL_LG) || (size != msg->size))
		{
			ret	= -EMSGSIZE;
			break;
		}

		switch (msg->status)
		{
			default:
			case OS_MSG_STS_MSG_TYPE:

				ret	= -EBADMSG;
				break;

			case OS_MSG_STS_OK:
			case OS_MSG_STS_MEM_MAP:
			case OS_MSG_STS_MSG_SIZE:

				ret	= msg->u.ioctl_lg.ret;
				break;

			case OS_MSG_STS_THREAD:

				ret	= -ENOEXEC;
				break;
		}

		break;
	}

	os_mem_handle_destroy(&mem);
	return(ret);
}



//*****************************************************************************
// Return Values:
//	0  = success
//	<0 = failure, value is -errno or -INtime error code
//	>0 = never returned
static int _ioctl_small(os_file_t** fp, const os_mbox_t* mbox, os_msg_t* msg, int cmd, void* buf, int buf_size)
{
	int			ret			= 0;
	int			msg_size	= OS_MSG_SIZE_IOCTL_SM_MIN + buf_size;
	WORD		size;
	BOOLEAN		sts;

	for (;;)	// A convenience loop.
	{
		// Initialize and send the message.
		memset(msg, 0, sizeof(os_msg_t));

		if (cmd & OS_IOCTL_DIR_WRITE)
			memcpy(msg->u.ioctl_sm.buf, buf, buf_size);

		msg->status			= OS_MSG_STS_OK;
		msg->msg_type		= OS_MSG_TYPE_IOCTL_SM;
		msg->size			= msg_size;
		msg->mbox			= mbox->mbox;
		msg->u.ioctl_sm.cmd	= cmd;
		sts					= SendRtData(fp[0]->dev_mbox, msg, msg_size);

		if (sts == FALSE)
		{
			ret	= os_get_error_status();
			break;
		}

		// Release the file pointer and don't use it after this point.
		os_file_unlock(fp);

		// Get and process the response.
		size	= ReceiveRtData(mbox->mbox, msg, WAIT_FOREVER);

		if (size == 0)
		{
			// The operation failed.
			ret	= os_get_error_status();
			break;
		}

		if (size != msg_size)
		{
			ret	= -EIO;
			break;
		}

		if (cmd & OS_IOCTL_DIR_READ)
			memcpy(buf, msg->u.ioctl_sm.buf, buf_size);

		switch (msg->status)
		{
			default:
			case OS_MSG_STS_MSG_TYPE:

				ret	= -EBADMSG;
				break;

			case OS_MSG_STS_OK:
			case OS_MSG_STS_MEM_MAP:
			case OS_MSG_STS_MSG_SIZE:

				ret	= msg->u.ioctl_sm.ret;
				break;

			case OS_MSG_STS_THREAD:

				ret	= -ENOEXEC;
				break;
		}

		break;
	}

	return(ret);
}



//*****************************************************************************
// Return Values:
//	0  = success
//	<0 = failure, value is -errno or -INtime error code
//	>0 = never returned
int os_api_ioctl(int fd, int cmd, void* arg)
{
	os_file_t*	fp		= NULL;
	os_mbox_t*	mbox	= NULL;
	os_msg_t*	msg		= NULL;
	int			ret;
	int			size	= OS_IOCTL_SIZE_DECODE(cmd);

	for (;;)	// a convenience loop
	{
		if (os_api_global.initialized == 0)
		{
			// The library has not been initialized.
			ret	= -EPROTO;
			break;
		}

		// Acquire file access.
		ret	= os_file_lock(fd, &fp);

		if (ret)
		{
			// There was a problem!
			break;
		}

		// Begin the IOCTL operation.

		// Acquire a mailbox for the receive message.
		ret	= os_mbox_alloc(&mbox);

		if (ret)
			break;

		// Allocate a message structure.
		ret	= os_msg_alloc(&msg);

		if (ret)
			break;

		if (size < 0)
		{
			ret	= -EINVAL;
		}
		else if (size <= sizeof(msg->u.ioctl_sm.buf))
		{
			ret	= _ioctl_small(&fp, mbox, msg, cmd, arg, size);
		}
		else
		{
			ret	= _ioctl_large(&fp, mbox, msg, cmd, arg, size);
		}

		break;
	}

	os_msg_free(&msg);
	os_mbox_free(&mbox);
	os_file_unlock(&fp);
	return(ret);
}


