// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AIO/16AIO_Linux_4.x.x.x_DN/driver/read.c $
// $Rev: 43208 $
// $Date: 2018-07-08 17:20:44 -0500 (Sun, 08 Jul 2018) $

// 16AIO: Device Driver: source file

#include "main.h"



/******************************************************************************
*
*	Function:	_get_fill_level
*
*	Purpose:
*
*		Determin the fill level of a FIFO. We do this via a binary search since
*		the threshold flag dynamically adjusts to the threshold value and the
*		FIFO fill level.
*
*		NOTE: The threshold flag may not respond appropriately if the FIFO is
*				full (32768 values) when the threshold level set to 0x7FFF
*				(32767). This routine therefore returns a value of 32767 even
*				when the FIFO is full.
*
*	Arguments:
*
*		dev		The structure for the device to access.
*
*	Returned:
*
*		>= 0	This is the nunmber of values in the FIFO and the transfer
*				buffer.
*
******************************************************************************/

static long _get_fill_level(dev_data_t* dev)
{
	u32		icr;
	u32		max	= 0x7FFF;
	u32		mid;
	u32		min	= 0x0000;
	u32		new;
	u32		obcr;
	long	qty	= 0;
	VADDR_T	vaddr	= dev->vaddr.gsc_ibcr_32;

	// Fix the interrupt option while we use the threshold value.
	icr		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_icr_32);

	if (((icr & 0x700) == 0x100) || ((icr & 0x700) == 0x200))
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, 0x700, 0x000);

	// Record the original threshold value;
	obcr	= os_reg_mem_rx_u32(dev, vaddr);

	// Do a binary search to get the buffer's fill level.
	for (;;)
	{
		mid	= (min + max) / 2;
		os_reg_mem_tx_u32(dev, vaddr, mid);
		new	= os_reg_mem_rx_u32(dev, vaddr);

		if (new & 0x10000)
		{
			// The level exceeds the value of mid.
			min	= mid + 1;

			if (min > max)
			{
				qty	= min;
				break;
			}
		}
		else
		{
			// The level meets or falls below the value of mid.

			if (min == max)
			{
				qty	= min;
				break;
			}

			max	= mid;
		}
	}

	// Restore the original threshold value;
	os_reg_mem_tx_u32(dev, vaddr, obcr);

	// Restore the interrupt option.

	if (((icr & 0x700) == 0x100) || ((icr & 0x700) == 0x200))
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, 0x700, icr);

	return(qty);
}



/******************************************************************************
*
*	Function:	dev_read_startup
*
*	Purpose:
*
*		Perform any work or tests needed before initiating a read operation.
*
*	Arguments:
*
*		alt		The device data structure.
*
*	Returned:
*
*		0		All is well.
*		< 0		There was an error.
*
******************************************************************************/

long dev_read_startup(GSC_ALT_STRUCT_T* alt)
{
	return(0);
}



/******************************************************************************
*
*	Function:	pio_read_available
*
*	Purpose:
*
*		Check to see how many bytes are available for reading.
*
*	Arguments:
*
*		dev		The device data structure.
*
*		count	What remains of the number of bytes that the application
*				requested.
*
*	Returned:
*
*		>= 0	The number of bytes available for reading.
*		< 0		There was an error.
*
******************************************************************************/

long pio_read_available(dev_data_t* dev, size_t count)
{
	long	qty;

	qty	= _get_fill_level(dev);
	qty	*= 4;
	return(qty);
}



/******************************************************************************
*
*	Function:	dma_read_available
*
*	Purpose:
*
*		Check to see how many bytes are available for reading.
*
*	Arguments:
*
*		dev		The device data structure.
*
*		count	What remains of the number of bytes that the application
*				requested.
*
*	Returned:
*
*		>= 0	The number of bytes available for reading.
*		< 0		There was an error.
*
******************************************************************************/

long dma_read_available(dev_data_t* dev, size_t count)
{
	u32	ibcr;
	u32	qty;

	ibcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_ibcr_32);

	if (ibcr & 0x10000)
	{
		// The fill level is above the threshold.
		qty	= (ibcr & 0x7FFF) * 4;
	}
	else
	{
		// The input buffer has too little data.
		qty	= 0;
	}

	return(qty);
}



/******************************************************************************
*
*	Function:	dma_read_work
*
*	Purpose:
*
*		Perform a read of the specified number of bytes.
*
*	Arguments:
*
*		dev		The device data structure.
*
*		mem		Put the data here.
*
*		count	What remains of the number of bytes that the application
*				requested.
*
*		st_end	The system timer count at which we timeout. If this is zero,
*				then we ignore this.
*
*	Returned:
*
*		>= 0	The number of bytes read.
*		< 0		There was an error.
*
******************************************************************************/

long dma_read_work(
	dev_data_t*		dev,
	const os_mem_t*	mem,
	size_t			count,
	os_time_tick_t	st_end)
{
	u32		dpr;
	u32		mode;
	long	qty;
	long	samples	= count / 4;

	if (samples < dev->rx.pio_threshold)
	{
		qty	= gsc_read_pio_work_32_bit(dev, mem, count, st_end);
	}
	else
	{
		mode	= GSC_DMA_MODE_BLOCK_DMA
				| GSC_DMA_MODE_SIZE_32_BITS
				| GSC_DMA_MODE_INPUT_ENABLE
				| GSC_DMA_MODE_BURSTING_LOCAL
				| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
				| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
				| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE;

		dpr		= GSC_DMA_DPR_BOARD_TO_HOST
				| GSC_DMA_DPR_END_OF_CHAIN
				| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty		= gsc_dma_perform(	dev,
									&dev->rx,
									st_end,
									GSC_DMA_CAP_DMA_READ,
									mode,
									dpr,
									mem,
									count);
	}

	return(qty);
}



/******************************************************************************
*
*	Function:	dmdma_read_available
*
*	Purpose:
*
*		Check to see how many bytes are available for reading.
*
*	Arguments:
*
*		dev		The device data structure.
*
*		count	What remains of the number of bytes that the application
*				requested.
*
*	Returned:
*
*		>= 0	The number of bytes available for reading.
*		< 0		There was an error.
*
******************************************************************************/

long dmdma_read_available(dev_data_t* dev, size_t count)
{
	// DMDMA is unsupported.
	return(-1);
}



/******************************************************************************
*
*	Function:	dmdma_read_work
*
*	Purpose:
*
*		Perform a read of the specified number of bytes.
*
*	Arguments:
*
*		dev		The device data structure.
*
*		mem		Put the data here.
*
*		count	What remains of the number of bytes that the application
*				requested.
*
*		st_end	The system timer count at which we timeout. If this is zero,
*				then we ignore this.
*
*	Returned:
*
*		>= 0	The number of bytes read.
*		< 0		There was an error.
*
******************************************************************************/

long dmdma_read_work(
	dev_data_t*		dev,
	const os_mem_t*	mem,
	size_t			count,
	os_time_tick_t	st_end)
{
	// DMDMA is unsupported.
	return(-1);
}


